#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <dlfcn.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glext.h>
#include <GL/glxext.h>
#include <X11/Xlib.h>
#include <X11/extensions/sync.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <pthread.h>
#include <bits/libc-lock.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <elf.h>
#include <link.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <error.h>
#include <sys/prctl.h>
#include <string.h>

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

static void* init_next_dlsym(void* handle, const char* symbol);
static typeof(dlsym)* NEXT_dlsym = init_next_dlsym;

static void* init_next_dlsym(void* handle, const char* symbol)
{
	struct link_map* lmap = (struct link_map*)dlopen("libdl.so.2", RTLD_LAZY);
	if(lmap)
	{
		char* base = (char*)lmap->l_addr;
		typeof(dlsym)* elf_dlsym = 0;
		const ElfW(Dyn)* dyn = 0;
		const ElfW(Sym) *symtab = 0;
		const Elf_Symndx *hash = 0;
		unsigned nbuckets;
		const char *strtab = 0;
		unsigned symhash = 0x6b3afd; /* elf hash of dlsym */
		unsigned symidx;

		dyn = lmap->l_ld;

		for(; dyn->d_tag; ++dyn)
		{
			char* ptr = (char*)dyn->d_un.d_ptr;
	#if defined(__mips__) || defined(__mips64__)
			ptr += base;
	#endif
			if(dyn->d_tag == DT_SYMTAB)
				symtab = (const ElfW(Sym)*)ptr;
			if(dyn->d_tag == DT_STRTAB)
				strtab = (const char*)ptr;
			if(dyn->d_tag == DT_HASH)
				hash = (const Elf_Symndx*)ptr;
		}
		if(!symtab || !strtab || !hash)
			goto error;

		nbuckets = *hash++;
		++hash;
		for(symidx = hash[symhash % nbuckets]; symidx != STN_UNDEF; symidx = hash[nbuckets + symidx])
		{
			const ElfW(Sym) *sym = &symtab[symidx];

			if(strcmp (strtab + sym->st_name, "dlsym"))
				continue;
			
			elf_dlsym = (typeof(dlsym)*)(base + sym->st_value);
			break;
		}

		if(!elf_dlsym)
			goto error;
		NEXT_dlsym = (typeof(dlsym)*)elf_dlsym(RTLD_NEXT, "dlsym");
		if(!NEXT_dlsym)
			goto error;

		return NEXT_dlsym(handle, symbol);
	}
	
  error:
	fprintf(stderr, "Fatal error: unable to find dlsym in libdl.so.2\n");
	abort();
	return 0;
}

static void* libx11;
static void* libgl;
static void* libglu;

void* next_dlsym(const char* symbol) __attribute__((visibility("hidden")));
 
void* next_dlsym(const char* symbol)
{
	void* ret;
	void* handle = 0;
	ret = NEXT_dlsym(RTLD_NEXT, symbol);
	if(!ret)
	{
		if(symbol[0] == 'X')
		{
			if(!libx11)
				libx11 = dlopen("libX11.so.6", RTLD_LAZY);
			handle = libx11;
		}
		else if((symbol[0] == 'g') && (symbol[1] == 'l'))
		{
			if(symbol[2] == 'u')
			{
				if(!libglu)
					libglu = dlopen("libGLU.so.1", RTLD_LAZY);
				handle = libglu;
			}
			else
			{
				if(!libgl)
					libgl = dlopen("libGL.so.1", RTLD_LAZY);
				handle = libgl;
			}
		}
	}
	if(handle)
		ret = NEXT_dlsym(handle, symbol);
	if(!ret)
	{
		fprintf(stderr, "Fatal error: unable to find symbol \"%s\".\n", symbol);
		abort();
	}
	return ret;
}

static void* self_handle;

static void* self_dlsym(const char* symbol)
{
	if(!self_handle)
	{
		Dl_info info;
		dladdr(&self_handle, &info);

		self_handle = dlopen(info.dli_fname, RTLD_LAZY | RTLD_NOLOAD);

//		printf("Name: %s Handle: %08x\n", info.dli_fname, self_handle);
		//self_handle = info.dli_fbase;
	}

//	printf("sym: %s\n", symbol);
	
	return NEXT_dlsym(self_handle, symbol);
}

__attribute__((visibility("default")))
void* __dlsym(void* handle, const char* symbol)
{
	void* ret;
	ret = self_dlsym(symbol);
	if(ret)
		return ret;
	else
		return NEXT_dlsym(handle, symbol);
}

asm(".symver __dlsym, dlsym@@GLIBC_2.0");

