#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xlibint.h>
#include <dlfcn.h>
#include <stdlib.h>
#include "ldpreload.h"

#define NEXT(func) ({static typeof(func)* NEXT_##func; if(!NEXT_##func) NEXT_##func = (typeof(func)*)NEXT_dlsym(RTLD_NEXT, #func); NEXT_##func;})

struct x11_display_info
{
	Display* dpy;
	Atom _NET_WM_USER_TIME;
	Atom _NET_WM_STATE;
	Atom _NET_WM_STATE_SKIP_TASKBAR;
	struct x11_display_info* next;
};

static struct x11_display_info* display_list;

static struct x11_display_info* get_display(Display* dpy);

static Status visibility_notify_wire_to_event(Display* dpy, XEvent* re, xEvent* event)
{
	struct x11_display_info* info = get_display(dpy);

	memset(re, 0, sizeof(*re));
	re->xvisibility.type = VisibilityNotify;
	re->xvisibility.display = dpy;
	re->xvisibility.window = event->u.visibility.window; 
	re->xvisibility.state = VisibilityUnobscured;
	return True;
}

static struct x11_display_info* get_display(Display* dpy)
{
	struct x11_display_info* p;
	for(p = display_list; p; p = p->next)
	{
		if(p->dpy == dpy)
			return p;
	}

	p= malloc(sizeof(*p));
	p->dpy = dpy;
	p->_NET_WM_USER_TIME = XInternAtom(dpy, "_NET_WM_USER_TIME", 0);
	p->_NET_WM_STATE = XInternAtom(dpy, "_NET_WM_STATE", 0);
	p->_NET_WM_STATE_SKIP_TASKBAR = XInternAtom(dpy, "_NET_WM_STATE_SKIP_TASKBAR", 0);

	XESetWireToEvent(dpy, VisibilityNotify, visibility_notify_wire_to_event);

	p->next = display_list;
	display_list = p;
}

static void before_map(Display *dpy, Window win)
{
	struct x11_display_info* info = get_display(dpy);
	unsigned user_time = 0;
	Atom skip_taskbar = info->_NET_WM_STATE_SKIP_TASKBAR;

	XChangeProperty(dpy, win, info->_NET_WM_USER_TIME, XA_CARDINAL, 32, PropModeReplace, (unsigned char*)&user_time, 1);
	XChangeProperty(dpy, win, info->_NET_WM_STATE, XA_ATOM, 32, PropModeReplace, (unsigned char*)&skip_taskbar, 1);
}

static void after_map(Display *dpy, Window win)
{
	XEvent event;

	memset(&event, 0, sizeof(event));
	event.xvisibility.type = VisibilityNotify;
	event.xvisibility.send_event = True; 
	event.xvisibility.display = dpy;
	event.xvisibility.window = win; 
	event.xvisibility.state = VisibilityUnobscured;
	XSendEvent(dpy,win,False,0,&event);

	memset(&event, 0, sizeof(event));
	event.xexpose.type = Expose;
	event.xexpose.send_event = True; 
	event.xexpose.display = dpy;
	event.xexpose.window = win; 
	event.xexpose.x = 0; event.xexpose.y = 0; 
	event.xexpose.width = 2048; event.xexpose.height = 2048;
	event.xexpose.count = 0;
	XSendEvent(dpy,win,False,0,&event);
}

int XMapWindow(Display *dpy, Window win)
{
	int ret;
	before_map(dpy, win);
	ret = NEXT(XMapWindow)(dpy, win);
	after_map(dpy, win);
	return ret;
}

int XMapRaised(Display *dpy, Window win)
{
	int ret;
	before_map(dpy, win);
	ret = NEXT(XMapWindow)(dpy, win);
	after_map(dpy, win);
	return ret;
}

int XMapSubwindows(Display *dpy, Window win)
{
	int ret;
	before_map(dpy, win);
	ret = NEXT(XMapSubwindows)(dpy, win);
	after_map(dpy, win);
	return ret;
}

int XRaiseWindow(Display *dpy, Window win)
{
	return 0;
}

